=begin
Copyright 2012 (c) TIG
All Rights Reserved.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###

TextureRandomizer.rb place it in the ../Plugins/ folder

###
Usage: TextureRandomizer

Used for randomizing textured materials like timber-grain on selected plank 
faces etc...

Preselect at least ONE face that has a front material with a texture.
Note if a selection contains other things they are ignored.

Activate the Tool:-
  'TextureRandomizer' in the Plugins menu.
OR type in the Ruby Console
  TextureRandomizer.new
plus <enter>

From v1.2...
If there are any groups/component-instances in the selection, and they 
contain textured-faces a dialog asks if you want to include those faces 
too, answer 'No' to process selected faces only or 'Yes' to process all 
textured-faces, including those inside other selected objects...
It does not 'mine' down beyond the selected context - so nested groups are 
never affected.

From v1.3...
If there are non-unique groups in the accepted selection then you are asked 
if you want to make them unique, so that their randomized-textures are also 
unique... [groups should be 'unique', BUT Sketchup can have more than one 
instance of a group !!]
If there are multiple instances of the same component in the accepted 
selection then you are asked if you want to make them unique, so that their 
randomized-textures are also unique... [use this with care as the advantage 
of components - edit one, edit all - is lost by making instances unique!]
  
A dialog appears with the following initial defaults:

Position UX max: 0
Position UX min: 0
Position VY max: 0
Position VY max: 0
These numbers are in 'current units' - e.g. 100, 1cm or 1".
They can be +ve or -ve; min=-50mm to max=50mm will nudge randomly between 
these two extremes; such small values to nudge textures 'across the grain' 
will probably work best for timber-grain textures; larger nudges along the 
grain might be more effective - experiment...
Using 0 for all values does not re-position the texture at all.

Scale UX max: 1
Scale UX min: 1
Scale VY max: 1
Scale VY min: 1
These numbers represent the scale-factor, so 1.1 is +10%.
They can be +ve or -ve; using a -ve value will 'mirror' all textures on 
that axis.
Using 1 for all values does not scale the texture at all.
Any value used must be more that 0.01 or less than -0.01, a value of 0 will 
be adjusted to 0.01; please remember that a tiny texture will probably be 
unrecognizable; and although there is no 'upper limit' a scaling-factor 
greater than ~100 would probably make the texture unrecognizably large !
Using small difference ranges [e.g. min=0.9 max=1.1] will probably give the 
best results for timber-grain etc.

Random Mirroring UX: No
Random Mirroring VY: No
Choose either 'Yes' or 'No'...
With 'No' then there is no 'mirroring' on that axis.
With 'Yes' the texture will be randomly 'mirrored' about that axis using 
the scaling-factors set above; so with '1' set for all scaling and 'Yes' 
you'll get 'mirroring' for ~50% of the selected faces' textures, randomly.

Rotation max: 0
Rotation min: 0
The Rotation is in degrees.
The values can be +ve or -ve.
Using 0 for both angles does not rotate the texture at all.
Any value > 360 will revert to an equivalent value below 360.
Any value < -360 will revert to an equivalent value above -360.
Small ranges like max=1 and min=-1 will probably give the best results for 
timber-grain etc.

All faces' material textures are now randomized to the parameters.

It is a 'one step' undoable change.

The previously used parameters are reused for the tool's dialog if it is 
activated later in that same session.

###
Donations:
info @ revitrev.org via PayPal.com
###
Versions:
1.0 20120320 First release.
1.1 20120320 Stupid typo fixed.
1.2 20120320 Option to include textured-faces inside selected 
             groups/components.
1.3 20120320 Non-unique groups and instances are now spotted, and you are 
             asked if you want to change them.
###
=end
#--------------------------------------------------------------------------
require 'sketchup.rb'
#--------------------------------------------------------------------------

module TextureRandomizer

# menu
unless file_loaded?(File.basename(__FILE__))
    menu=UI.menu("Plugins")
    cmd1=UI::Command.new("TextureRandomizer"){TextureRandomizer.new()}
	cmd1.tooltip="TextureRandomizer"
	cmd1.status_bar_text="TextureRandomizer: Randomize Position/Scale/Rotation of a Selected Faces' Textures..."
	cmd1.menu_text="TextureRandomizer"
	menu.add_item(cmd1)
    ###
end
file_loaded(File.basename(__FILE__))

# code
def self.new()
  @model=Sketchup.active_model
  @model.start_operation("TextureRandomizer")
  @ss=@model.selection
  @faces=@ss.to_a.find_all{|e|e.class==Sketchup::Face and e.material and e.material.texture}
  conts=@ss.to_a.find_all{|e|e.class==Sketchup::Group or e.class==Sketchup::ComponentInstance}
  cfaces=[]
  conts.each{|c|
    ents=c.entities if c.class==Sketchup::Group
    ents=c.definition.entities if c.class==Sketchup::ComponentInstance
    ents.each{|e|
      cfaces << e if e.class==Sketchup::Face and e.material and e.material.texture
    }
  }
  cfaces.flatten!
  if cfaces[0] ### 6=YES 7=NO
    UI.beep
    res1=UI.messagebox("TextureRandomizer:\n\nInclude textured-faces inside the selected groups/components ?",MB_YESNO,"")
    if res1==6
      ### groups
      groups=@ss.to_a.find_all{|e|e.class==Sketchup::Group}
      nonuniqs=groups.find_all{|g|self.definition(g).instances[1]}
      nonuniqGtexs=[]
      nonuniqs.each{|g|
        g.entities.each{|e|
          if e.class==Sketchup::Face and e.material and e.material.texture
            nonuniqGtexs << g
          end
        }
      }
      nonuniqGtexs.uniq!
      if nonuniqGtexs[0]
        UI.beep
        res2=UI.messagebox("TextureRandomizer:\n\nMake selected groups unique ?",MB_YESNO,"")
        gfaces=[]
        nonuniqs.each{|g|
          g.make_unique if self.definition(g).instances[1]
          g.entities.each{|e|
            if e.class==Sketchup::Face and e.material and e.material.texture
              gfaces << e
            end
          }
        } if res2==6
        cfaces=cfaces+gfaces
        cfaces.uniq!
      end
      ### instances
      instances=@ss.to_a.find_all{|e|e.class==Sketchup::ComponentInstance}
      nonuniqs=instances.find_all{|e|e.definition.instances[1]}
      nonuniqCtexs=[]
      nonuniqs.each{|i|
        i.definition.entities.each{|e|
          if e.class==Sketchup::Face and e.material and e.material.texture
            nonuniqCtexs << i
          end
        }
      }
      nonuniqCtexs.uniq!
      if nonuniqCtexs[0]
        UI.beep
        res3=UI.messagebox("TextureRandomizer:\n\nMake selected component-instances unique ?",MB_YESNO,"")
        ifaces=[]
        nonuniqs.each{|i|
          i.make_unique if i.definition.instances[1]
          i.definition.entities.each{|e|
            if e.class==Sketchup::Face and e.material and e.material.texture
              ifaces << e
            end
          }
        } if res3==6
        cfaces=cfaces+ifaces
        cfaces.uniq!
      end
      ###
      @faces= @faces+cfaces
      ###
    end
  end#if
  ###
  @faces.uniq!
  @faces.compact!
  ###
  unless @faces[0]
    UI.messagebox("TextureRandomizer:\n\nSelect at least one face with a textured material!")
    @model.abort_operation
    return nil
  end
  ###
  unless self.dialog()
    @model.abort_operation
    return nil
  end
  ###
  self.process()
  ###
  @model.commit_operation
  ###
end#def

def self.dialog()
  @pxmax='0'.to_l unless @pxmax
  @pxmin='0'.to_l unless @pxmin
  @pymax='0'.to_l unless @pymax
  @pymin='0'.to_l unless @pymin
  ###
  @sxmax=1.0 unless @sxmax
  @sxmin=1.0 unless @sxmin
  @symax=1.0 unless @symax
  @symin=1.0 unless @symin
  ###
  @mirrx='No' unless @mirrx
  @mirry='No' unless @mirry
  ###
  @rmax=0.0 unless @rmax
  @rmin=0.0 unless @rmin
  prompts=['Position UX max: ','Position UX min: ','Position VY max: ','Position VY min: ','Scale UX max: ','Scale UX min: ','Scale VY max: ','Scale VY min: ','Random Mirroring UX: ','Random Mirroring VY:','Rotation max: ','Rotation min: ']
  values=[@pxmax,@pxmin,@pymax,@pymin,@sxmax,@sxmin,@symax,@symin,@mirrx,@mirry,@rmax,@rmin]
  pops=['','','','', '','','','', 'Yes|No','Yes|No', '','']
  title='TextureRandomizer: Parameters'
  results=inputbox(prompts,values,pops,title)
  return nil unless results
  @pxmax,@pxmin,@pymax,@pymin,@sxmax,@sxmin,@symax,@symin,@mirrx,@mirry,@rmax,@rmin = results
  ### reset if 'bad'
  @sxmax=0.01 if @sxmax>=0 and @sxmax<0.01
  @sxmin=0.01 if @sxmin>=0 and @sxmin<0.01
  @symax=0.01 if @symax>=0 and @symax<0.01
  @symin=0.01 if @symin>=0 and @symin<0.01
  @sxmax=-0.01 if @sxmax<0 and @sxmax>-0.01
  @sxmin=-0.01 if @sxmin<0 and @sxmin>-0.01
  @symax=-0.01 if @symax<0 and @symax>-0.01
  @symin=-0.01 if @symin<0 and @symin>-0.01
  @rmax=@rmax -360 while @rmax>= 360
  @rmax=@rmax +360 while @rmax<= -360
  @rmin=@rmin -360 while @rmin>= 360
  @rmin=@rmin +360 while @rmin<= -360
  ###
  return true
end#def

def self.process()
 ### Position
 @faces.each{|face|
  ###
  if @pxmax<@pxmin
    pxmax,pxmin=@pxmin,@pxmax
  else
    pxmax,pxmin=@pxmax,@pxmin
  end
  if @pymax<@pymin
    pymax,pymin=@pymin,@pymax
  else
    pymax,pymin=@pymax,@pymin
  end
  ###
  u = pxmin+((pxmax-pxmin)*rand)
  v = pymin+((pymax-pymin)*rand)
  ###
  mat=face.material
  tw=Sketchup.create_texture_writer
  samples=[face.outer_loop.vertices[0].position]  ### 0,0 | Origin
  samples << samples[0].offset(face.normal.axes.x)### 1,0 | Offset Origin in X
  samples << samples[1].offset(face.normal.axes.y)### 1,1 | Offset X in Y
  samples << samples[0].offset(face.normal.axes.y)### 0,1 | Offset Origin in Y
  pts=[]### Array containing 3D points.
  uvs=[]### Array containing UV points.
  uvh=face.get_UVHelper(true, false, tw)
  samples.each{|pt|
    pts << pt                                      ### XYZ 3D coordinates 
    uvs << self.normalizeUVQ(uvh.get_front_UVQ(pt))### UVQ 2D coordinates
  }
  vx=Geom::Vector3d.new(u,0,0)
  vy=Geom::Vector3d.new(0,v,0)
  tt=Geom::Transformation.new(pts[0], face.normal)
  vx.transform!(tt)
  vy.transform!(tt)
  tx=Geom::Transformation.translation(vx)
  ty=Geom::Transformation.translation(vy)
  tr=tx * ty
  ###
  tpts=[]
  (0..3).each{|i|
    tpts << pts[i].transform(tr)
    tpts << uvs[i]
  }
  face.position_material(mat, tpts, true)
 } unless @pxmax==0 and @pxmin==0 and @pymax==0 and @pymin==0
 ### Scale
 @faces.each{|face|
  ###
  if @sxmax<@sxmin
    sxmax,sxmin=@sxmin,@sxmax
  else
    sxmax,sxmin=@sxmax,@sxmin
  end
  if @symax<@symin
    symax,symin=@symin,@symax
  else
    symax,symin=@symax,@symin
  end
  ###
  u = sxmin+((sxmax-sxmin)*rand)
  v = symin+((symax-symin)*rand)
  ###
  mat=face.material
  tw=Sketchup.create_texture_writer
  samples=[face.outer_loop.vertices[0].position]  ### 0,0 | Origin
  samples << samples[0].offset(face.normal.axes.x)### 1,0 | Offset Origin in X
  samples << samples[1].offset(face.normal.axes.y)### 1,1 | Offset X in Y
  samples << samples[0].offset(face.normal.axes.y)### 0,1 | Offset Origin in Y
  pts=[]### Array containing 3D points.
  uvs=[]### Array containing UV points.
  uvh=face.get_UVHelper(true, false, tw)
  samples.each{|pt|
    pts << pt                                      ### XYZ 3D coordinates 
    uvs << self.normalizeUVQ(uvh.get_front_UVQ(pt))### UVQ 2D coordinates
  }
  vx=Geom::Vector3d.new(u,0,0)
  vy=Geom::Vector3d.new(0,v,0)
  tt=Geom::Transformation.new(pts[0], face.normal)
  vx.transform!(tt)
  vy.transform!(tt)
  ###
  tpts=[pts[0], uvs[0]]
  (1..3).each{|i|
    case i
      when 1
        pt=pts[0].offset(vx)
      when 2
        pt=pts[0].offset(vx)
        pt.offset!(vy)
      when 3
        pt=pts[0].offset(vy)
    end
    tpts << pt
    tpts << uvs[i]
  }
  face.position_material(mat, tpts, true)
 } unless @sxmax==1 and @sxmin==1 and @symax==1 and @symin==1
  ### Mirrored
 @faces.each{|face|
  if rand(2)==1
    u =  1.0
  else
    u = -1.0
  end
  if rand(2)==1
    v =  1.0
  else
    v = -1.0
  end
  u=1.0 if @mirrx=='No'
  v=1.0 if @mirry=='No'
  mat=face.material
  tw=Sketchup.create_texture_writer
  samples=[face.outer_loop.vertices[0].position]  ### 0,0 | Origin
  samples << samples[0].offset(face.normal.axes.x)### 1,0 | Offset Origin in X
  samples << samples[1].offset(face.normal.axes.y)### 1,1 | Offset X in Y
  samples << samples[0].offset(face.normal.axes.y)### 0,1 | Offset Origin in Y
  pts=[]### Array containing 3D points.
  uvs=[]### Array containing UV points.
  uvh=face.get_UVHelper(true, false, tw)
  samples.each{|pt|
    pts << pt                                      ### XYZ 3D coordinates 
    uvs << self.normalizeUVQ(uvh.get_front_UVQ(pt))### UVQ 2D coordinates
  }
  vx=Geom::Vector3d.new(u,0,0)
  vy=Geom::Vector3d.new(0,v,0)
  tt=Geom::Transformation.new(pts[0], face.normal)
  vx.transform!(tt)
  vy.transform!(tt)
  ###
  tpts=[pts[0], uvs[0]]
  (1..3).each{|i|
    case i
      when 1
        pt=pts[0].offset(vx)
      when 2
        pt=pts[0].offset(vx)
        pt.offset!(vy)
      when 3
        pt=pts[0].offset(vy)
    end
    tpts << pt
    tpts << uvs[i]
  }
  face.position_material(mat, tpts, true)
 } if @mirrx=='Yes' or @mirry=='Yes'
  ### Rotation
 @faces.each{|face|
  ###
  if @rmax<@rmin
    rmax,rmin=@sxmin,@rmax
  else
    rmax,rmin=@rmax,@rmin
  end
  angle = rmin+((rmax-rmin)*rand)
  ###
  mat=face.material
  tw=Sketchup.create_texture_writer
  samples=[face.outer_loop.vertices[0].position]  ### 0,0 | Origin
  samples << samples[0].offset(face.normal.axes.x)### 1,0 | Offset Origin in X
  samples << samples[1].offset(face.normal.axes.y)### 1,1 | Offset X in Y
  samples << samples[0].offset(face.normal.axes.y)### 0,1 | Offset Origin in Y
  pts=[]### Array containing 3D points.
  uvs=[]### Array containing UV points.
  uvh=face.get_UVHelper(true, false, tw)
  samples.each{|pt|
    pts << pt                                      ### XYZ 3D coordinates 
    uvs << self.normalizeUVQ(uvh.get_front_UVQ(pt))### UVQ 2D coordinates
  }
  tr=Geom::Transformation.rotation(pts[0], face.normal, angle.degrees)
  tpts=[pts[0], uvs[0]]
  (1..3).each{|i|
    tpts << pts[i].transform(tr)
    tpts << uvs[i]
  }
  face.position_material(mat, tpts, true)
 } unless @rmax==0 and @rmin==0
  ###
end#def

def self.normalizeUVQ(uvq)
  uvq[0] /= uvq[2]
  uvq[1] /= uvq[2]
  uvq[2] = 1.0
  return uvq
end#def

def self.definition(group)
  @model.definitions.each{|defn|
    return defn if defn.instances.include?(group)
  }
  return nil ### Should never happen.
end#def

end#module
